"""
Cleans and tokenizes given text
"""
import re


def clean(text):
    """
    :type text str
    :rtype: str
    """
    # basic formatting
    text = re.sub(r"'''?([^']+)'''?", '\\1', text)

    # templates
    # {{foo}}
    # {{foo|{{test}}|123}}
    while '{{' in text:
        start = text.find('{{')
        level = 1
        pos = start + 2

        while pos < len(text):
            # print('>' + text[pos:pos+2] + '<')

            if text[pos:pos+2] == '{{':
                # nested template - enter next level
                level += 1
                pos += 1
            elif text[pos:pos+2] == '}}':
                # nested template - leave this level
                pos += 1
                level -= 1

            # template is now completed
            if level == 0:
                # print(text, start, pos, text[start:pos+1])
                text = text[:start] + ' ' + text[pos+1:]
                break

            # check next character
            pos += 1

        # the template is not well balanced, leave the endless loop
        if level != 0:
            break

    # tables
    text = re.sub(r'{\|[^}]+\|}', '', text)  # {|foo..|}

    # headings
    text = re.sub(r'^=+\s?([^=]+)\s?=+', lambda matches: matches.group(1).strip(),
                  text, flags=re.MULTILINE)  # == a == -> a

    # files and other links with namespaces
    text = re.sub(r'\[\[[^:\]]+:[^\]]+\]\]', '', text)  # [[foo:b]] -> ''

    # local links
    text = re.sub(r'\[\[([^|\]]+)\]\]', '\\1', text)  # [[a]] -> a
    text = re.sub(r'\[\[[^|]+\|([^\]]+)\]\]', '\\1', text)  # [[a|b]] -> b

    text = text.replace('[[', '').replace(']]', '')

    # external links
    text = re.sub(r'\[http[^\s]+ ([^\]]+)\]', '\\1', text)  # [[http://example.com foo]] -> foo
    text = re.sub(r'https?://[^\s]+', '', text)  # remove http://example.com

    # lists
    text = re.sub(r'^\*+\s?', '', text, flags=re.MULTILINE)

    # parser hooks
    text = re.sub(r'<[^>]+>[^<]+</[^>]+>', ' ', text)  # <ref>foo</ref>

    # HTML
    text = re.sub(r'<[^>]+/?>', ' ', text)  # <br> / <br />
    text = text.replace('&nbsp;', ' ')

    return text.strip()


def tokenize_filter(text):
    """
    :type text str
    :rtype: bool
    """
    if text == '':
        return False

    if text[0].isdigit() and re.fullmatch(r'\d+', text):
        return False

    return True


def tokenize(text, filter_func=tokenize_filter):
    """
    :type text str
    :type filter_func callable
    :rtype: list[str]
    """
    # clean up the text
    text = re.sub(r'[?.,:;!()=+"]', ' ', text)  # remove noise

    text = text.strip()

    # tokenize
    parts = re.split(r'[-–\s/|_&{}\xAD]', text)

    parts = filter(filter_func, parts)

    return list(parts)



"""
Class used to parse provided XML dump and emit a stream of articles content

https://gist.github.com/macbre/1543d945f5244c5c68681966f07e2d6c
"""
import logging

from xml import sax


class DumpHandler(sax.ContentHandler):
    """
    SAX ContentHandler instance of Wikipedia context XML dumps

      <page>
        <title>Wikipedia:All pages by title</title>
        <ns>4</ns>
        <id>4</id>
        <restrictions>sysop</restrictions>
        <revision>
          <id>6498</id>
          <timestamp>2004-05-25T02:19:28Z</timestamp>
          <contributor>
            <ip>82.68.206.22</ip>
          </contributor>
          <model>wikitext</model>
          <format>text/x-wiki</format>
          <text xml:space="preserve">foo bar</text>
          <sha1>edilrfbz8rbsr9mvl8dflarkw56ojy5</sha1>
        </revision>
      </page>
    """

    # https://www.mediawiki.org/wiki/Help:Export#Export_format
    # https://docs.python.org/3.6/library/xml.sax.handler.html#xml.sax.handler.ContentHandler

    # pylint: disable=too-many-instance-attributes
    def __init__(self):
        super(DumpHandler, self).__init__()
        self.logger = logging.getLogger(self.__class__.__name__)

        self.pages_batch = []
        self.pages_count = 0

        # parser state, are we inside <page> or <revision> tag?
        self.in_page = False
        self.in_revision = False
        self.in_contributor = False

        # concatenated content of tags
        self.tag_content = ''

        # currently parsed page dump
        self.current_title = ''
        self.current_namespace = 0
        self.current_page_id = 0
        self.current_revision_id = 0
        self.current_revision_timestamp = 0
        self.current_content = ''

    def reset_state(self):
        """
        Reset page information
        """
        self.in_page = False
        self.in_revision = False
        self.in_contributor = False

        self.current_title = ''
        self.current_namespace = 0
        self.current_page_id = 0
        self.current_revision_id = 0
        self.current_revision_timestamp = 0
        self.current_content = ''

    def startElement(self, name, attrs):
        # print('>', name, attrs)
        # self.logger.info('> startElement %s %s', name, attrs)

        if name == 'page':
            self.in_page = True
        elif name == 'revision':
            self.in_revision = True
        elif name == 'contributor':
            self.in_contributor = True

        self.tag_content = ''

    def endElement(self, name):
        # print('<', name, self.tag_content)
        # self.logger.info('< endElement %s (%s)', name, self.tag_content)

        if name == 'page':
            self.in_page = False

            # add next page information
            self.logger.debug('Page #%d: %s', self.current_page_id, self.current_title)

            self.pages_batch.append((
                self.current_namespace,
                self.current_page_id,
                self.current_title,
                self.current_content,
                self.current_revision_id,
                self.current_revision_timestamp,
            ))

            self.pages_count += 1

            self.reset_state()
            return
        if name == 'revision':
            self.in_revision = False
            return
        if name == 'contributor':
            self.in_contributor = False
            return

        if self.in_contributor:
            pass
        elif self.in_revision:
            if name == 'id':
                self.current_revision_id = int(self.tag_content)
            elif name == 'timestamp':
                self.current_revision_timestamp = self.tag_content
            elif name == 'text':
                self.current_content = self.tag_content
        elif self.in_page:
            if name == 'title':
                self.current_title = self.tag_content
            elif name == 'ns':
                self.current_namespace = int(self.tag_content)
            elif name == 'id':
                self.current_page_id = int(self.tag_content)

    def characters(self, content):
        # print('=', content)
        # self.logger.info('= characters %s', content)

        self.tag_content += content

    def get_pages(self):
        """
        Used by DumpReader to yield pages as we parse the XML dump
        """
        for page in self.pages_batch:
            yield page

        self.pages_batch = []

    def get_pages_count(self):
        """
        :rtype: int
        """
        return self.pages_count


class DumpReader:
    """
    This class uses provided BaseDump instance to read and parse MediaWiki's XML dump
    """
    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)

    @staticmethod
    def filter_by_namespace(namespace):
        """
        :type namespace int
        :rtype bool
        """
        return isinstance(namespace, int)

    def read(self, dump):
        """
        :type dump mediawiki_dump.dumps.BaseDump
        :rtype: list
        """
        self.logger.info('Parsing XML dump...')

        # https://docs.python.org/2/library/xml.etree.elementtree.html#parsing-xml

        # https: // docs.python.org / 3.6 / library / xml.sax.html?highlight = sax  # xml.sax.parse
        handler = DumpHandler()

        parser = sax.make_parser()
        parser.setContentHandler(handler)

        for chunk in dump.get_content():
            parser.feed(chunk)

            # yield pages as we go through XML stream
            for page in handler.get_pages():
                (namespace, page_id, title, content, revision_id, revision_timestamp) = page

                if content == '':
                    # https://fo.wikipedia.org/wiki/Kjak:L%C3%ADvfr%C3%B8%C3%B0i
                    self.logger.warning('Page #%d: %s is empty', page_id, title)
                    continue

                if self.filter_by_namespace(namespace):
                    yield namespace, page_id, title, content, revision_id, revision_timestamp

        self.logger.info('Parsing completed, pages found: %d', handler.get_pages_count())


class DumpReaderArticles(DumpReader):
    """
    Use this class to get article pages only
    """
    @staticmethod
    def filter_by_namespace(namespace):
        """
        :type namespace int
        :rtype bool
        """
        return namespace == 0



from setuptools import setup, find_packages

VERSION = '0.2'

# @see https://github.com/pypa/sampleproject/blob/master/setup.py
setup(
    name='mediawiki_dump',
    version=VERSION,
    author='Maciej Brencz',
    author_email='maciej.brencz@gmail.com',
    license='MIT',
    description='Python package for working with MediaWiki XML content dumps',
    url='https://github.com/macbre/mediawiki_dump',
    # https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Topic :: Text Processing :: Markup :: XML',

        # Pick your license as you wish
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    packages=find_packages(),
    extras_require={
        'dev': [
            'coverage==4.5.1',
            'pylint==2.1.1',
            'pytest==3.9.2',
        ]
    },
    install_requires=[
        'libarchive-c==2.8',
        'requests==2.20.0',
    ]
)



"""
CLasses that support fetching dumps
"""
import bz2
import logging

from hashlib import md5
from os.path import isfile
from tempfile import gettempdir

import libarchive
import requests
from requests.exceptions import HTTPError


class DumpError(Exception):
    """
    Generic exception class
    """
    pass


class BaseDump:
    """
    A generic dump class. Wikipedia or Wikia dumps
    should have a separate class that customizes get_url() method
    """
    ARCHIVE_FORMAT = 'bz2'

    def __init__(self, wiki):
        """
        :type wiki str
        """
        self.wiki = wiki
        self.logger = logging.getLogger(self.__class__.__name__)

        self.http = requests.session()
        self.http.headers['User-Agent'] = \
            'python-mediawiki-dump (+https://github.com/macbre/mediawiki-dump)'

    def get_cache_filename(self, url):
        """
        Return a hashed filename of cache entry for a given URL

        :type url str
        :rtype: str
        """
        _hash = md5()
        _hash.update(url.encode('utf-8'))

        return 'mediawiki_dump_{hash}.{extension}'.format(
            hash=_hash.hexdigest(), extension=self.ARCHIVE_FORMAT)

    def get_url(self):
        """
        :rtype: str
        """
        raise NotImplementedError('fetch method needs to be implemented')

    def fetch(self):
        """
        :rtype: _io.TextIOWrapper
        """
        url = self.get_url()

        cache_filename = '{}/{}'.format(gettempdir(), self.get_cache_filename(url))
        self.logger.info("Checking %s cache file...", cache_filename)

        # check cache
        if not isfile(cache_filename):
            # fetch the resource
            self.logger.info('Fetching %s dump from <%s>...', self.wiki, url)
            response = self.http.get(url, stream=True)
            self.logger.info('HTTP %s (%d kB fetched)',
                             response.status_code, len(response.content) / 1024)

            # raise an exception and do not set a cache entry
            try:
                response.raise_for_status()
            except HTTPError as ex:
                self.logger.error('Failed to fetch a dump', exc_info=True)
                raise DumpError('Failed to fetch a dump, request ended with HTTP {}'.
                                format(ex.response.status_code))

            # read the response as a stream and put it into cache file
            # http://docs.python-requests.org/en/master/user/advanced/#body-content-workflow
            #
            # before using a stream reading and parsing of Faroese dump made the words_from_dump.py
            # script took ~460 MB of memory, after the change - ~60 MB
            with response:
                with open(cache_filename, 'wb') as file:
                    for chunk in response.iter_content():
                        file.write(chunk)

                response.close()
                self.logger.info("Cache set")
        else:
            self.logger.info("Reading from cache")

        # return a stream of compressed data from the cache file
        return open(cache_filename, 'rb')

    def get_content(self):
        """
        :rtype: list[str]
        """
        raise NotImplementedError('fetch method needs to be implemented')


class WikipediaDump(BaseDump):
    """
    Class for fetching Wikipedia dumps from https://dumps.wikimedia.org
    """
    def get_url(self):
        return 'https://dumps.wikimedia.org/{wiki}/latest/' \
               '{wiki}-latest-pages-meta-current.xml.bz2'.format(wiki='{}wiki'.format(self.wiki))

    def get_content(self):
        """
        :rtype: list[str]
        """
        # https://docs.python.org/3.6/library/bz2.html#bz2.BZ2Decompressor
        decompressor = bz2.BZ2Decompressor()

        with self.fetch() as content:
            for chunk in content:
                yield decompressor.decompress(chunk)


class WikiaDump(BaseDump):
    """
    Class for fetching Wikia dumps

    https://community.wikia.com/wiki/Help:Database_download
    """
    ARCHIVE_FORMAT = '7z'

    def get_url(self):
        # https://muppet.wikia.com/wiki/Special:Statistics
        return 'https://s3.amazonaws.com/wikia_xml_dumps/{}/{}/{}_pages_current.xml.7z'.format(
            self.wiki[0], self.wiki[:2], self.wiki)

    def get_content(self):
        """
        :rtype: list[str]
        """
        # https://github.com/Changaco/python-libarchive-c#usage
        with self.fetch() as handler:
            with libarchive.file_reader(handler.name) as archive:
                for entry in archive:
                    for block in entry.get_blocks():
                        yield block
